-module(oam_config).
-author('tobbe@eddieware.org').
%%%----------------------------------------------------------------------
%%% File    : oam_config.erl
%%% Created : 12 Jul 1999 by tobbe@eddieware.org
%%% Function: Routines for manipulating the configuration data.
%%%----------------------------------------------------------------------
%%%
%%% The contents of this file are subject to the Erlang Public License,
%%% Version 1.0, (the "License"); you may not use this file except in
%%% compliance with the License. You may obtain a copy of the License at
%%% http://www.eddieware.org/EPL
%%%
%%% The contents of this file are subject to the Erlang Public License
%%% License, Version 1.0, (the "License"); you may not use this file
%%% except in compliance with the License. You may obtain a copy of the
%%% License at http://www.eddieware.org/EPL
%%%
%%% Software distributed under the License is distributed on an "AS IS"
%%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%%% the License for the specific language governing rights and limitations
%%% under the License.
%%%
%%% The Original Code is Eddie-0.83b1.
%%%
%%% The Initial Developer of the Original Code is Ericsson Telecom
%%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%%% Telecom AB. All Rights Reserved.
%%%
%%% Contributor(s): ______________________________________.
%%%
%%%----------------------------------------------------------------------
-vc('$Id: oam_config.erl,v 1.19 1999/09/06 05:07:32 tobbe Exp $ ').

-export([get_dns_servers/0,set_dns_servers/1,get_netmask/0,set_netmask/1,
	 add_new_node/1,get_node_interfaces/1,set_node_interfaces/2,
	 get_node_load_threshold/1,set_node_load_threshold/2,get_frontends/0,
	 add_new_cluster/1,get_cluster/1,add_new_template/1,get_node/1,
	 get_templates/1,get_template/2,set_admit_ctrl/1,get_admit_ctrl/1,
	 add_admit_static/1,get_admit_static/1,set_frontend/1,get_clusters/0,
	 get_frontend/1,add_erlets/1,get_erlets/1,add_backend/1,get_erlets/0,
	 get_backends/0,get_backends/1,get_backends/2,set_cluster_type/2,
	 get_admit_never/0,get_admit_always/0,get_node/0,get_backends_node/1]).

-include("db.hrl").
-include("inet_server.hrl").

-define( I(X), integer(X) ).
-define( IP(A,B,C,D), 
	 ?I(A),?I(B),?I(C),?I(D),A>=0,A<256,B>=0,B<256,C>=0,C<256,D>=0,D<256 ).


%%% --------------------------------------------------------------------
%%% Operations towards the root record.
%%% --------------------------------------------------------------------

get_dns_servers() -> 
    case g_root() of
	{atomic,[R]} -> {ok,R#root.dns_servers};
	{_,Reason}   -> {error,Reason}
    end.

set_dns_servers(Servers) when list(Servers) -> 
    dns_servers_format(Servers),
    F = fun() -> 
		[R] = mnesia:read({root,main}),
		mnesia:write(R#root{dns_servers=Servers})
	end,
    mnesia_exec(F).

%%% Format should be a list of: {HostName,PortNo}

dns_servers_format([]) -> true;
dns_servers_format([{Host,Port}|T]) when list(Host),integer(Port) ->
    dns_servers_format(T);
dns_servers_format([H|_]) ->
    exit({error,{wrong_format,H}}).

get_netmask() -> 
    case g_root() of
	{atomic,[R]} -> {ok,R#root.netmask};
	{_,Reason}   -> {error,Reason}
    end.

set_netmask({A,B,C,D}) when ?I(A),?I(B),?I(C),?I(D) -> 
    F = fun() -> 
		[R] = mnesia:read({root,main}),
		mnesia:write(R#root{netmask={A,B,C,D}})
	end,
    mnesia_exec(F).

g_root() ->
    F = fun() -> mnesia:read({root,main}) end,
    mnesia:transaction(F).

%%% --------------------------------------------------------------------
%%% Operations towards the node record
%%% ----------------------------------
%%% Just for those curious people out there: I think that the 
%%% Interfaces directive really should be named FailoverInterfaces 
%%% to make it clear what it is used for (i.e if an interface doesn't
%%% work Eddie will try another alternative).
%%% --------------------------------------------------------------------

add_new_node(N) when record(N,node),atom(N#node.name),
		     list(N#node.interfaces),
		     number(N#node.load_threshold) ->
    mnesia_write(N).

%%% Get all nodes 

get_node() ->
    W = mnesia:table_info(node,wild_pattern),
    F = fun() -> mnesia:match_object(W)	end,
    case mnesia:transaction(F) of
	{atomic,Nodes}   -> {ok,Nodes};
	{aborted,Reason} -> {error,Reason}
    end.

%%% Get a particular node

get_node(Name) when atom(Name) -> 
    case g_node(Name) of
	{atomic,[Node]}  -> {ok,Node};
	{aborted,Reason} -> {error,Reason}
    end.

get_node_interfaces(Name) when atom(Name) ->
    case g_node(Name) of
	{atomic,[N]} when record(N,node) -> 
	    {ok,N#node.interfaces};
	{_,Reason} ->
	    {error,Reason}
    end.

set_node_interfaces(Name,If) when atom(Name),list(If) ->
    F = fun() -> 
		case mnesia:read({node,Name}) of
		    [N] -> mnesia:write(N#node{interfaces=If});
		    _   -> mnesia:abort({node_not_found,Name})
		end
	end,
    mnesia_exec(F).
    
get_node_load_threshold(Name) when atom(Name) ->
    case g_node(Name) of
	{atomic,[N]} when record(N,node) -> 
	    {ok,N#node.load_threshold};
	{_,Reason} ->
	    {error,Reason}
    end.

set_node_load_threshold(Name,Load) when atom(Name),integer(Load) ->
    F = fun() -> 
		case mnesia:read({node,Name}) of
		    [N] -> mnesia:write(N#node{load_threshold = Load});
		    _   -> mnesia:abort({node_not_found,Name})
		end
	end,
    mnesia_exec(F).
    
g_node(Name) ->
    F = fun() -> mnesia:read({node,Name}) end,
    mnesia:transaction(F).

%%% --------------------------------------------------------------------
%%% Operations towards the cluster record.
%%% --------------------------------------------------------------------

add_new_cluster(C) when record(C,cluster),list(C#cluster.name),
			atom(C#cluster.cluster_type),
			list(C#cluster.failover_nodes),
			list(C#cluster.backend_clusters),
			list(C#cluster.ip_address_pool) ->
    check_cluster_type(C#cluster.cluster_type),
    check_ip_address_pool(C#cluster.ip_address_pool),
    check_failover_nodes(C),
    check_backend_clusters(C),
    mnesia_write(C).

get_cluster(ClusterName) when list(ClusterName) ->
    case g_cluster(ClusterName) of
	{atomic,[C]} -> {ok,C};
	_ -> throw({error,{non_existing_cluster,ClusterName}})
    end.

%%% Get all the clusters
get_clusters() ->
    F = fun() -> 
		W = mnesia:table_info(cluster,wild_pattern),
		mnesia:match_object(W)
	end,
    case mnesia:transaction(F) of
	{atomic,Clusters} -> {ok,Clusters};
	{_,Reason}        -> {error,Reason}
    end.

check_cluster_type(frontend) -> true;
check_cluster_type(backend)  -> true;
check_cluster_type(Type)     -> 
    exit({error,{wrong_cluster_type,Type}}).
    
check_ip_address_pool([]) -> true;
check_ip_address_pool([{A,B,C,D}|T]) when ?I(A),?I(B),?I(C),?I(D) ->
    check_ip_address_pool(T);
check_ip_address_pool([H|_]) -> 
    exit({error,{wrong_ip_address_format,H}}).
    
%%% Check that each node in the failover list, really is defined.

check_failover_nodes(C) ->
    F = fun(NodeName) ->
		case g_node(NodeName) of
		    {atomic,_} -> ok;
		    _ -> throw({error,{non_existing_failover_node,
				       NodeName}})
		end
	end,
    case catch lists:foreach(F,C#cluster.failover_nodes) of
	{error,Reason} -> exit({error,Reason});
	_ -> true
    end.
		
%%% Check that each cluster in specified as backend clusters
%%% really is defined.

check_backend_clusters(C) when C#cluster.cluster_type == frontend ->
    F = fun(ClusterName) ->
		case g_cluster(ClusterName) of
		    {atomic,_} -> ok;
		    _ -> throw({error,{non_existing_backend_cluster,
				       ClusterName}})
		end
	end,
    case catch lists:foreach(F,C#cluster.backend_clusters) of
	{error,Reason} -> exit({error,Reason});
	_ -> true
    end;
check_backend_clusters(C) ->
    %% Hey, since we apparently aren't a frontend we
    %% can't specify backend clusters, can we ?
    exit({error,{cannot_define_backend_clusters,C#cluster.name}}).
    
set_cluster_type(Name,frontend) when list(Name) -> 
    set_cluster_type_0(Name,frontend);
set_cluster_type(Name,backend) when list(Name) -> 
    set_cluster_type_0(Name,backend).

set_cluster_type_0(Name,Type) ->
    F = fun() -> 
		case mnesia:read({cluster,Name}) of
		    [C] -> mnesia:write(C#cluster{cluster_type = Type});
		    _   -> mnesia:abort({cluster_not_found,Name})
		end
	end,
    mnesia_exec(F).

g_cluster(Name) ->
    F = fun() -> mnesia:read({cluster,Name}) end,
    mnesia:transaction(F).


%%% --------------------------------------------------------------------
%%% Operations towards the template record.
%%% --------------------------------------------------------------------

add_new_template(T) when record(T,template),list(T#template.interface),
			 ?I(T#template.port),T#template.port>=0,
			 ?I(T#template.max_retries) ->
    check_autoconfig(T),
    check_ip_address(T),
    check_start(T),
    check_stop(T),
    check_monitor(T),
    check_notify(T),
    F = fun() -> 
		case mnesia:read({template,T#template.cluster}) of
		    {atomic,Ts} ->
			mnesia:write(T#template{start_order=(length(Ts) + 1)});
		    _ ->
			mnesia:abort({error,failed_to_read_template_table})
		end
	end,
    mnesia_exec(F).
		
get_templates(ClusterName) when list(ClusterName) ->
    case g_templates(ClusterName) of
	{atomic,Ts} -> Ts;
	_ -> {error,no_templates_found}
    end.

get_template(ClusterName,NodeName) when list(ClusterName),atom(NodeName) ->
    W = mnesia:table_info(template,wild_pattern),
    F = fun() -> 
		mnesia:match_object(W#template{cluster=ClusterName,
					       node=NodeName}) 
	end,
    case mnesia:transaction(F) of
	{atomic,[T]} when record(T,template) -> {ok,T};
	{atomic,Ts} when length(Ts)>1 -> 
	    %% NB: This shouldn't happen ! We include this
	    %% case only for debugging purposes...
	    exit({error,{identical_nodes_in_cluster,Ts}});
	Else ->
	    {error,no_template_found}
    end.

check_autoconfig(T) when T#template.auto_config == on  -> true;
check_autoconfig(T) when T#template.auto_config == off -> true;
check_autoconfig(T) -> 
    exit({error,{wrong_autoconfig_value,T#template.auto_config}}).

check_ip_address(T) when T#template.ip_address == dynamic -> true;
check_ip_address(T) when tuple(T#template.ip_address),
			 size(T#template.ip_address) == 4 ->
    case T#template.ip_address of
	{A,B,C,D} when ?IP(A,B,C,D) -> true;
	Else ->
	    exit({error,{wrong_ip_address_format,Else}})
    end;
check_ip_address(T) ->
    exit({error,{wrong_ip_address_format,T#template.ip_address}}).

check_start(T) when list(T#template.start) ->
    check_scripts(T#template.start).

check_stop(T) when list(T#template.stop) ->
    check_scripts(T#template.stop).

check_monitor(T) when list(T#template.monitor) ->
    check_scripts(T#template.monitor).

check_notify(T) when list(T#template.notify) ->
    check_scripts(T#template.notify).

check_scripts(List) when list(List) ->
    F = fun({Node,module,Mod,Fun,Args}) when atom(Node),atom(Mod),
					atom(Fun),list(Args) -> true;
	   ({Node,exec,Exec,Args}) when atom(Node),list(Exec),
					list(Args) -> true;
	   (Else) ->
		exit({error,{wrong_script_format,Else}})
	end,
    lists:foreach(F,List).

g_templates(ClusterName) when list(ClusterName) ->
    F = fun() -> mnesia:read({template,ClusterName}) end,
    mnesia:transaction(F).


%%% --------------------------------------------------------------------
%%% Operations towards the admit_ctrl record.
%%% --------------------------------------------------------------------

set_admit_ctrl(A) when record(A,admit_ctrl),
		       list(A#admit_ctrl.frontend),
		       ?I(A#admit_ctrl.time),
		       list(A#admit_ctrl.blocked_page),
		       list(A#admit_ctrl.reject_page),
		       list(A#admit_ctrl.queue_page),
		       ?I(A#admit_ctrl.queue_places),
		       ?I(A#admit_ctrl.max_sessions) ->
    mnesia_write(A).

get_admit_ctrl(FrontEnd) when list(FrontEnd) ->
    F = fun() -> mnesia:read({admit_ctrl,FrontEnd}) end,
    case mnesia:transaction(F) of
	{atomic,[A]} -> {ok,A};
	_ -> {error,not_found}
    end.

%%% --------------------------------------------------------------------
%%% Operations towards the admit_static record.
%%% --------------------------------------------------------------------

add_admit_static(A) when record(A,admit_static),
			 tuple(A#admit_static.ip),
			 size(A#admit_static.ip) == 4 ->
    check_status(A),
    mnesia_write(A).

get_admit_never()   -> get_admit_static(never).
get_admit_always()  -> get_admit_static(always).

get_admit_static(never)  -> get_admit_static_0(never);
get_admit_static(always) -> get_admit_static_0(always).

get_admit_static_0(Status) -> 
    W = mnesia:table_info(admit_static,wild_pattern),
    F = fun() -> mnesia:match_object(W#admit_static{status=Status}) end,
    case mnesia:transaction(F) of
	{atomic,As} -> {ok,As};
	_           -> {error,not_found}
    end.

    
check_status(A) when A#admit_static.status == never -> true;
check_status(A) when A#admit_static.status == always -> true;
check_status(A) -> 
    exit({error,{wrong_status,A#admit_static.status}}).

%%% --------------------------------------------------------------------
%%% Operations towards the frontend record.
%%% --------------------------------------------------------------------

%%% NB: Tri ! The time values should be in milliseconds !!

set_frontend(Fe) when record(Fe,frontend),list(Fe#frontend.name),
		      atom(Fe#frontend.protocol_module),
		      list(Fe#frontend.endpoints_config) ->
    check_endpoints_config(Fe#frontend.endpoints_config),
    check_admit_ctrl(Fe),
    mnesia_write(Fe).

get_frontends() ->
    F = fun() -> 
		W = mnesia:table_info(frontend,wild_pattern),
		mnesia:match_object(W) 
	end,
    case mnesia:transaction(F) of
	{atomic,Fes} -> {ok,Fes};
	{_,Reason}   -> {error,Reason}
    end.

get_frontend(Name) when list(Name) ->
    F = fun() -> mnesia:read({frontend,Name}) end,
    case mnesia:transaction(F) of
	{atomic,[Fe]} -> {ok,Fe};
	_ -> {error,not_found}
    end.

check_endpoints_config([]) -> true;
check_endpoints_config([E|Es]) when record(E,endpoint_config),
				    atom(E#endpoint_config.name),
				    ?I(E#endpoint_config.port), 
				    E#endpoint_config.port >= 0,
				    ?I(E#endpoint_config.read_timeout),
				    E#endpoint_config.read_timeout >= 0,
				    ?I(E#endpoint_config.keep_alive_timeout),
				    E#endpoint_config.keep_alive_timeout >= 0 ->
    check_endpoints_config(Es);
check_endpoints_config([E|_]) ->
    exit({error,{endpoint_config,E}}).

check_admit_ctrl(Fe) when Fe#frontend.admit_ctrl == true  -> true;
check_admit_ctrl(Fe) when Fe#frontend.admit_ctrl == false -> true;
check_admit_ctrl(Fe) ->
    exit({error,{admit_ctrl_not_boolean,Fe#frontend.admit_ctrl}}).

%%% --------------------------------------------------------------------
%%% Operations towards the backend record.
%%% --------------------------------------------------------------------

add_backend(B) when record(B,backend),atom(B#backend.endpoint_name),
		    atom(B#backend.backend_node) ->
    check_spec(B#backend.spec),
    check_schedule_pattern(B#backend.schedule_patterns),
    mnesia_write(B).

%%% Note that the 'backend' table is of type 'bag'.
%%% This means that a key may return several values.
%%% The (default) key is EndpointName = 'http_tcp' which
%%% also is the only possible key value so far (Aug-1999).
%%% To restrict outcome get_backends/2 gives the possibility
%%% to also specify a particula node. This "should" result
%%% in a unique backend record being returned.

get_backends() -> get_backends(http_tcp).

get_backends(EPname) when atom(EPname) ->
    W = mnesia:table_info(backend,wild_pattern),
    get_backends_0(W#backend{endpoint_name=EPname}).

get_backends_node(BeNode) when atom(BeNode) ->
    W = mnesia:table_info(backend,wild_pattern),
    get_backends_0(W#backend{backend_node=BeNode}).

get_backends(EPname,BeNode) when atom(EPname),atom(BeNode) ->
    W = mnesia:table_info(backend,wild_pattern),
    get_backends_0(W#backend{endpoint_name=EPname,
			     backend_node=BeNode}).

get_backends_0(W) when record(W,backend) ->
    F = fun() -> mnesia:match_object(W) end,
    case mnesia:transaction(F) of
	{atomic,Result} -> {ok,Result};
	_               -> {error,not_found}
    end.

check_spec({proxy,{A,B,C,D},Port}) when ?IP(A,B,C,D),?I(Port) -> true;
check_spec({erlets,Name}) when atom(Name)                     -> true;
check_spec(X) -> exit({error,{backend_sched_spec,X}}).

check_schedule_pattern({Key,{Patterns,AntiPatterns}}) when atom(Key),
							   list(Patterns),
							   list(AntiPatterns) ->
    true;
check_schedule_pattern(Else) ->
    exit({error,{wrong_schedule_pattern,Else}}).
    
%%% --------------------------------------------------------------------
%%% Operations towards the erlets record.
%%% NB: As it is today (12 Aug 1999) no #gen_storage{} records
%%%     are ever generated since no 'Erlet' directives exist.
%%%     So we don't bother implement operations for it.
%%% --------------------------------------------------------------------

add_erlets(E) when record(E,erlets),atom(E#erlets.name),list(E#erlets.list) ->
    mnesia_write(E).

get_erlets() ->
    F = fun() ->
		W = mnesia:table_info(erlets,wild_pattern),
		mnesia:match_object(W)
	end,
    case mnesia:transaction(F) of
	{atomic,Es}      -> {ok,Es};
	{aborted,Reason} -> {error,Reason}
    end.

get_erlets(Name) when atom(Name) ->
    F = fun() -> mnesia:read({erlets,Name}) end,
    case mnesia:transaction(F) of
	{atomic,[E]} -> {ok,E};
	_            -> {error,not_found}
    end.

%%% --------------
%%% Misc. routines

mnesia_write(Data) ->
    F = fun() -> mnesia:write(Data) end,
    mnesia_exec(F).

mnesia_exec(F) when function(F) ->
    case mnesia:transaction(F) of
	{atomic,_} -> ok;
	{_,Reason} -> {error,Reason}
    end.


