-module(oam_config_files).
-author('tobbe@eddieware.org').
%%%----------------------------------------------------------------------
%%% File    : oam_config_files.erl
%%% Created : 16 Aug 1999 by tobbe@eddieware.org
%%% Function: Routines for writing config files.
%%%----------------------------------------------------------------------
%%%
%%% The contents of this file are subject to the Erlang Public License,
%%% Version 1.0, (the "License"); you may not use this file except in
%%% compliance with the License. You may obtain a copy of the License at
%%% http://www.eddieware.org/EPL
%%%
%%% The contents of this file are subject to the Erlang Public License
%%% License, Version 1.0, (the "License"); you may not use this file
%%% except in compliance with the License. You may obtain a copy of the
%%% License at http://www.eddieware.org/EPL
%%%
%%% Software distributed under the License is distributed on an "AS IS"
%%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%%% the License for the specific language governing rights and limitations
%%% under the License.
%%%
%%% The Original Code is Eddie-0.83b1.
%%%
%%% The Initial Developer of the Original Code is Ericsson Telecom
%%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%%% Telecom AB. All Rights Reserved.
%%%
%%% Contributor(s): ______________________________________.
%%%
%%%----------------------------------------------------------------------
-vc('$Id: oam_config_files.erl,v 1.6 1999/09/05 15:16:00 geoff Exp $ ').

-export([write/1,write/2]).

-include("db.hrl").
-include("../../inet_server/src/inet_server.hrl").

%%% --------------------------------------------------------------------
%%% Write the config data to the corresponding .mig/.gate files
%%% --------------------------------------------------------------------

write(PrefixName) when list(PrefixName) ->
    {ok,CurDir} = file:get_cwd(),
    write(PrefixName,CurDir).

write(PrefixName,Dir) when list(PrefixName),list(Dir) ->
    write_mig_file(PrefixName,Dir),
    write_gate_file(PrefixName,Dir).

write_mig_file(PrefixName,Dir) -> 
    Fname = PrefixName ++ ".mig",
    MigFile = build_mig_file(),
    write_file(Dir ++ "/" ++ Fname,MigFile).

write_gate_file(PrefixName,Dir) -> 
    Fname = PrefixName ++ ".gate",
    GateFile = build_gate_file(),
    write_file(Dir ++ "/" ++ Fname,GateFile).

%%% Have to use file:write/2 here since we 
%%% are dealing with a deep I/O list.
write_file(Fname,File) ->
    case file:open(Fname,write) of
	{ok,Fd} ->
	    file:write(Fd,File),
	    file:sync(Fd),
	    file:close(Fd),
	    ok;
	Else ->
	    Else
    end.

%%% --------------------------------------------------------------------
%%% Build a .gate file
%%% --------------------------------------------------------------------

build_gate_file() ->
    %% Will we ever get more than one frontend here... ?
    [Fe] = oam_config:get_frontends(),
    ["<Frontend ",Fe#frontend.name,">\n\n",
     "  ProtocolModule ",Fe#frontend.protocol_module,"\n\n",
     build_endpoint_config(Fe),
     build_admit_info(Fe),
     build_erlets(),
     build_backends(Fe#endpoint_config.name)].


%%% The backend table is of type bag. That means that the key 
%%% 'endpoint_name' may result in a number of 'backend' records, 
%%% where each record represent a (endpoint_name,backend_node) 
%%% combination.  So to be able to build the schedule directive 
%%% we have to partition the set of records according to the 
%%% 'backend_node', and for each backend_node build the 
%%% corresponding Schedule directives.

build_backends(EPname) ->
    Bs = lists:keysort(#backend.backend_node,oam_config:get_backends(EPname)),
    build_backend(EPname,Bs).

build_backend(EPname,Bs) ->
    {Set,Rest} = first_node_partition(Bs),
    build_backend(EPname,Set,Rest).

build_backend(EPname,BeNodes,Bs) ->
    F = fun(B) -> 
		["  <Backend ",l2a(EPname)," ",B#backend.backend_node,">\n",
		 build_schedule(BeNodes),
		 "  </Backend>\n"]
	end,
    Be = lists:map(F,BeNodes),
    {Set,Rest} = first_node_partition(Bs),
    [Be|build_backend(EPname,Set,Rest)].

first_node_partition([H|T]) ->
    Node = H#backend.backend_node,
    first_node_partition(Node,T,[H]).

first_node_partition(_,[],Acc) -> 
    {lists:reverse(Acc),[]};
first_node_partition(Node,[H|T],Acc) when H#backend.backend_node == Node ->
    first_node_partition(Node,T,[H|Acc]);
first_node_partition(Node,Rest,Acc) ->
    {lists:reverse(Acc),Rest}.

%%% The #backend.schedule_patterns field contains tuples
%%% looking like: {Key, {Patterns, AntiPatterns}}

build_schedule(BeNodes) ->
    F = fun(B) ->
	   {Key,{Patterns,AntiPatterns}} = B#backend.schedule_patterns,
	   ["    <Schedule>\n",
	    "      Patterns ",Key," ",Patterns," ",anti(AntiPatterns),"\n",
	    build_schedule_type(B#backend.spec),
	    "    </Schedule>\n"]
	end,
    lists:map(F,BeNodes).

anti(Patterns) ->
    F = fun(P) -> [$^|P] end,
    lists:map(F,Patterns).

build_schedule_type({erlets,Erlets}) -> 
    ["      Erlets ",Erlets,"\n"];
build_schedule_type({proxy,Ip,Port}) -> 
    ["      Proxy ",ip2str(Ip)," ",i2l(Port),"\n"].

build_erlets() ->
    Es = oam_config:get_erlets(),
    F = fun(E) when record(E,erlets) ->
		["  \n<Erlets ",E#erlets.name,">\n",
		 build_erlets_list(E#erlets.list),
		 "  </Erlets>\n"]
	end,
    lists:map(F,Es).

build_erlets_list(L) when list(L) ->
    F = fun(E) ->
		["    <Erlet ",l2a(E),">\n",
		 "    </Erlet>\n"]
	end,
    lists:map(F,L).

build_admit_info(Fe) ->
    Ac = oam_config:get_admit_ctrl(Fe#frontend.name),
    Always = oam_config:get_admit_always(),
    Never = oam_config:get_admit_never(),
    ["  AdmitControl ",bool2str(Fe#frontend.admit_ctrl),"\n\n",
     "  AdmitTime ",Ac#admit_ctrl.time,"\n\n",
     "  AdmitBlockedPage ",Ac#admit_ctrl.blocked_page,"\n\n",
     "  AdmitRejectPage ",Ac#admit_ctrl.reject_page,"\n\n",
     "  AdmitQueuePage ",Ac#admit_ctrl.queue_page,"\n\n",
     "  AdmitQueuePlaces ",Ac#admit_ctrl.queue_page,"\n\n",
     "  AdmitMaxSessions ",Ac#admit_ctrl.max_sessions,"\n\n",
     "  AdmitAlways ",static2str(Always),"\n\n",
     "  AdmitNever ",static2str(Never),"\n\n"].
     
static2str(L) when list(L) ->
    F = fun(A) when record(A,admit_static) ->
		Ip = A#admit_static.ip,
		[ip2str(Ip)," "]
	end,
    lists:map(F,L).

bool2str(true)  -> "true";
bool2str(false) -> "false".

build_endpoint_config(Fe) ->
    F = fun(E) ->
		["  <EndpointConfig ",E#endpoint_config.name,">\n",
		 "    Port ",E#endpoint_config.port,"\n",
		 "    ReadTimeout ",i2l(E#endpoint_config.read_timeout),"\n",
		 "    KeepAliveTimeout ",
		 i2l(E#endpoint_config.keep_alive_timeout),"\n",
		 "  </EndpointConfig>\n\n"]
	end,
    lists:map(F,Fe).


%%% --------------------------------------------------------------------
%%% Build a .mig file
%%% --------------------------------------------------------------------

build_mig_file() ->
    DNS = build_dns_server(),
    Netmask = build_netmask(),
    Nodes = build_nodes(),
    Clusters = build_clusters(),
    [DNS,Netmask,Nodes,Clusters].

build_dns_server() ->
    case oam_config:get_dns_servers() of
	{ok,Ds} when list(Ds) ->
	    F = fun({Host,Port}) -> 
			"DNSServer " ++ Host ++ " " ++ Port ++ "\n"
		end,
	    lists:map(F,Ds);
	{error,Reason} ->
	    exit(Reason)
    end.
    
build_netmask() ->
    case oam_config:get_netmask() of
	{ok,{A,B,C,D}} ->
	    "Netmask " ++ ip2dotted(A,B,C,D) ++ "\n";
	{error,Reason} ->
	    exit(Reason)
    end.

build_nodes() ->
    {ok,Nodes} = oam_config:get_node(),
    F = fun(N) ->
		["<Node ",N#node.name,">\n",
		 "  Interfaces ",N#node.interfaces,"\n",
		 "  LoadThreshold 0 ",N#node.load_threshold,"\n",
		 "</Node>\n"]
	end,
    lists:map(F,Nodes).

build_clusters() ->
    {ok,Clusters} = oam_config:get_clusters(),
    F = fun(C) ->
		Servers = build_servers(C),
		["<Cluster ",C#cluster.name,">\n",
		 "  ClusterType ",cluster_type(C),"\n\n",
		 failover_nodes(C),
		 backend_clusters(C),
		 Servers,
		 "\n</Cluster>\n\n"]
	end,
    lists:map(F,Clusters).

build_servers(C) ->
    {ok,Templates} = oam_config:get_templates(C#cluster.name),
    F = fun(T) ->
		["\n<Server ",T#template.node,">\n",
		 auto_config(T),
		 "  Interface ",T#template.interface,"\n",
		 "  IPAddress ",ip2str(T#template.ip_address),"\n",
		 "  Port ",i2l(T#template.port),"\n",
		 "  Start ",script2str(T#template.start),"\n",
		 "  Stop ",script2str(T#template.stop),"\n",
		 "</Server>\n"]
	end,
    lists:map(F,Templates).

script2str({Node,module,M,F,A}) ->
    [a2l(Node)," module ",a2l(M)," ",a2l(F)," ",A];
script2str({Node,exec,E,A}) ->
    [a2l(Node)," exec ",E," ",A].

ip2str(dynamic)   -> "dynamic";
ip2str({A,B,C,D}) -> [i2l(A),".",i2l(B),".",i2l(C),".",i2l(D)].

auto_config(T) when T#template.auto_config == off -> "  AutoConfig Off\n";
auto_config(T) when T#template.auto_config == on  -> "  AutoConfig On\n".
		 
backend_clusters(C) when C#cluster.cluster_type == frontend ->
    ["  BackendClusters",C#cluster.backend_clusters,"\n"];
backend_clusters(C) -> [].

failover_nodes(C) when C#cluster.cluster_type == frontend ->
    ["  FailoverNodes",C#cluster.failover_nodes,"\n"];
failover_nodes(C) -> [].

cluster_type(C) when C#cluster.cluster_type == frontend -> "Frontend";
cluster_type(C) when C#cluster.cluster_type == backend  -> "Backend".
    
ip2dotted(A,B,C,D) ->
    l2i(A) ++ "." ++ l2i(B) ++ "." ++ l2i(C) ++ "." ++ l2i(D).

l2i(L) when list(L)    -> list_to_integer(L);
l2i(I) when integer(I) -> I.

i2l(I) when integer(I) -> integer_to_list(I);
i2l(L) when list(L)    -> L.

l2a(L) when list(L) -> list_to_atom(L);
l2a(A) when atom(A) -> A.

a2l(A) when atom(A) -> atom_to_list(A);
a2l(L) when list(L) -> L.


